#!/usr/bin/perl
#
# Guestbook
#
# This software is under Public Domain
#
# Requirements
# 	Perl 5.005 or later (maybe)
# 	CGI::Minimal (Tested with 1.29)
# 	CGI::FastTemplate (Tested with 1.09)
# 	Digest::MD5
# 	AtExit

use strict;
use warnings;

use lib qw(. lib); # library path
use vars qw(
$C_MAX_MESSAGE_SIZE $C_MAX_ATTACHMENT_SIZE $C_MIN_SIZE $C_MAX_MESSAGES
@C_KEEP_EXTENSIONS

$C_MAX_PAGE $C_MESSAGE_DIR $C_LISTFILE $C_LOCKDIR $C_LOCKFILE
$C_REBUILD_FILE $C_REBUILD_INTERVAL $C_THIS_URL $C_LIST_URL $C_MESSAGE_URL

$T_FORM $T_LIST $T_NAV_PREV $T_NAV_NEXT $T_NO_ITEM
$T_STATUS $T_NAV_LIST $T_MESSAGE $T_LIST_ITEM $T_STATUS_ITEM

$E_LOCK $E_MESSAGE_NOME $E_MESSAGE_SHORT $E_MESSAGE_LONG $E_MESSAGE_TEXT
$E_MESSAGE_SAME $E_OVERWRITE_NAME $E_OVERWRITE_SIZE
$E_REBUILD_FILE $E_REBUILD_INTERVAL $E_REBUILD_DIRECTORY
$E_REQUEST $E_WRITE_LIST $E_WRITE_MESSAGE

$P_MESSAGE $P_ATTACH $P_OVERWRITE $P_PAGE $P_REBUILD

$RE_FILENAME $RE_FILESIZE
);

# configulation
#

# Max message size to accept. Unit is byte.
$C_MAX_MESSAGE_SIZE = 8 * 1024;
# Max attachment size to accept. Unit is byte.
$C_MAX_ATTACHMENT_SIZE = 128 * 1024;
# Min message size to accept. Unit is byte.
$C_MIN_SIZE = 64;
# Max messages to save.
$C_MAX_MESSAGES = 128;
# Max messages shown on list page at once.
$C_MAX_PAGE = 32;
# Filename extensions of attachment to keep.
# First matched extention is used.
@C_KEEP_EXTENSIONS = qw(gif jpg jpeg png svg svgz bmp mid
                        7z lzh rar tar.gz tar gz zip);
# Directory to store messages.
# This directory must be writable (w attribute)
# and executable (x attribute) by this CGI.
$C_MESSAGE_DIR = 'messages';
# Filename for list. (with path)
# A name with '.tmp' suffix is also used.
# The directory that this file is in must be writable (w attribute)
# and executable (x attribute) by the CGI.
$C_LISTFILE = "$C_MESSAGE_DIR/guestbook.txt";
# Directory name of using lock.
# This directory must be readable (r attribute)
# and executable (x attribute) by the CGI.
$C_LOCKDIR = 'lock/';
# Filename for lock. (name only)
# This file must be readable (r attribute)
# and writable (w attribute) by the CGI.
$C_LOCKFILE = 'guestbook.lock';
# Filename to use rebuild. (with path)
# A name with '.tmp' suffix is also used.
# The directory that this file is in must be writable (w attribute)
# and executable (x attribute) by the CGI.
$C_REBUILD_FILE = "data/guestbook.rebuild";
# Interval to allow rebuilding. Unit is seconds.
$C_REBUILD_INTERVAL = 60 * 60;
# URL of this script.
# This is used to make a link refers to this script.
$C_THIS_URL = 'guestbook.cgi';
# URL of list file.
# This is shown on form page if you want it.
$C_LIST_URL = '';
# Base URL of messages.
# This is used to make a link refers to messages on the list page.
# This is shown on form page if you want it.
$C_MESSAGE_URL = '';

# templates
#
$T_FORM = <<'EOT';
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html
PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<title>Guestbook</title></head>
<body>
<p>A message or an attachment is required,
message name to overwrite is optional.
Minimum size of a message and an attachment is
<span id="min_size">$C_MIN_SIZE</span> bytes.
Maximum messages size is
<span id="max_message_size">$C_MAX_MESSAGE_SIZE</span> bytes.
Maximum attachments size is
<span id="max_attachment_size">$C_MAX_ATTACHMENT_SIZE</span> bytes.
Maximum number to stock messages and attachments is
<span id="max_messages">$C_MAX_MESSAGES</span>.</p>
<form method="post" action="$C_THIS_URL" enctype="multipart/form-data" accept-charset="utf-8">
<p>Message</p>
<p><textarea name="$P_MESSAGE" tabindex="1" rows="10" cols="60" accesskey="1"></textarea></p>
<p>Attachment</p>
<p><input type="file" name="$P_ATTACH" tabindex="2" size="60" maxlength="50" accesskey="2" /></p>
<p>Message name to overwrite (optional)</p>
<p><input type="text" name="$P_OVERWRITE" value="$OVERWRITE" tabindex="3" size="60" maxlength="50" accesskey="3" /></p>
<p><input type="submit" tabindex="4" name="submit" accesskey="3" />
<input type="reset" tabindex="5" name="reset" accesskey="4" /></p></form>
<p><a href="$C_THIS_URL?$P_PAGE=1">List messages.</a></p>
<p><a href="/">Back to main page.</a></p>
<div id="url_list">$C_LIST_URL</div>
<div id="url_pool">$C_MESSAGE_URL</div>
<div><span id="param_message">$P_MESSAGE</span>
<span id="param_attachment">$P_ATTACH</span>
<span id="param_overwrite">$P_OVERWRITE</span>
<span id="param_page">$P_PAGE</span>
<span id="param_rebuild">$P_REBUILD</span></div>
<div><span id="param_extensions">$EXTENSIONS</span></div>
<p><a href="$C_THIS_URL?$P_REBUILD=1">Rebuild messages list.</a></p>
</body>
</html>
EOT
$T_LIST = <<'EOT';
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html
PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<title>Guestbook</title></head>
<body>
$T_NAV_PREV
$T_LIST_ITEM
$T_NAV_NEXT
<p><a href="$C_THIS_URL">Back to form.</a></p>
<p><a href="/">Back to main page.</a></p>
<p><a href="$C_THIS_URL?$P_REBUILD=1">Rebuild messages list.</a></p>
</body>
</html>
EOT
$T_NAV_PREV = '<p><a id="nav_previous" href="$C_THIS_URL?$P_PAGE=$PAGE">previous</a></p>';
$T_NAV_NEXT = '<p><a id="nav_next" href="$C_THIS_URL?$P_PAGE=$PAGE">next</a></p>';
$T_NO_ITEM = '<div>No message.</div>';
$T_STATUS = <<'EOT';
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html
PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<title>Guestbook</title></head>
<body>
<p id="status">$STATUS</p>
$T_MESSAGE
<p><a href="$C_THIS_URL">Back to form.</a></p>
$T_NAV_LIST
<p><a href="/">Back to main page.</a></p>
</body>
</html>
EOT
$T_NAV_LIST = '<p><a href="$C_THIS_URL?$P_PAGE=1">List messages.</a></p>';
$T_MESSAGE = '<p class="reason">$REASON</p>';
$T_LIST_ITEM = '<div><a href="$C_THIS_URL?$P_OVERWRITE=$ITEM_NAME">overwrite</a> <a class="message" href="$C_MESSAGE_URL/$ITEM_NAME">$ITEM_NAME</a></div>';
$T_STATUS_ITEM = '<div><a class="message" href="$C_MESSAGE_URL/$ITEM_NAME">$ITEM_NAME</a></div>';

# error messages
#
$E_LOCK = "500 Can't lock.";
$E_MESSAGE_NOME = '400 Legal message is none.';
$E_MESSAGE_SHORT = '400 Message is too short.';
$E_MESSAGE_LONG = '400 Message is too long.';
$E_MESSAGE_TEXT = '400 Illegal message text.';
$E_MESSAGE_SAME = '400 Message is already posted.';
$E_OVERWRITE_NAME = '400 Illegal name for overwrite.';
$E_OVERWRITE_SIZE = '400 Message size is too small for overwrite.';
$E_REBUILD_FILE = "500 Can't write the file of rebuild information.";
$E_REBUILD_INTERVAL = "500 Can't rebuild list.";
$E_REBUILD_DIRECTORY = "500 Can't read messages.";
$E_REQUEST = '400 Invalid request.';
$E_WRITE_LIST = "500 Can't write the list.";
$E_WRITE_MESSAGE = "500 write error: ";

# command parameters
#
$P_MESSAGE = 'message';
$P_ATTACH = 'attach';
$P_OVERWRITE = 'overwrite';
$P_PAGE = 'page';
$P_REBUILD = 'rebuild';

# hashing library
#

# MD5
use Digest::MD5 qw(md5_hex);
#use Digest::Perl::MD5 qw(md5_hex);

# SHA
#use Digest::SHA qw(sha1_hex);
#use Digest::SHA::PurePerl qw(sha1_hex);
#use Digest::SHA qw(sha256_hex);
#use Digest::SHA::PurePerl qw(sha256_hex);

# filename
# 
$RE_FILENAME = '^[0-9A-Fa-f]+-\d+\.(?:'.
               join('|', 'txt', 'bin', map(quotemeta, @C_KEEP_EXTENSIONS)).
               ')$';
$RE_FILESIZE = '^[0-9A-Fa-f]+-(\d+)\.(?:'.
               join('|', 'txt', 'bin', map(quotemeta, @C_KEEP_EXTENSIONS)).
               ')$';
sub generate_filename {
	sprintf('%s-%d', md5_hex($_[0]), length($_[0]));
}

# constrain message text
#

# no check
sub check_message_text {
	1;
}

# check utf8.
# sub check_message_text {
# 	my $err = shift;
# 	use utf8;
# 	&check_error($err, utf8::valid($_[0]), $E_MESSAGE_TEXT);
# }

# check utf8 by regexp.
# sub check_message_text {
# 	my $err = shift;
# 	# RFC 3629
# 	my $utf8_regexp = "[\x00-\x7f]|[\xc2-\xdf][\x80-\xbf]|".
# 		"\xe0[\xa0-\xbf][\x80-\xbf]|[\xe1-\xec][\x80-\xbf]{2}|".
# 		"\xed[\x80-\x9f][\x80-\xbf]|[\xee-\xef][\x80-\xbf]{2}|".
# 		"\xf0[\x90-\xbf][\x80-\xbf]{2}|[\xf1-\xf3][\x80-\xbf]{3}|".
# 		"\xf4[\x80-\x8f][\x80-\xbf]{2}";
# 	&check_error($err, $_[0] =~ /^(?:$utf8_regexp){1,256}/og, $E_MESSAGE_TEXT)
# 		or return;
# 	my $p;
# 	while ($_[0] =~ /\G(?:$utf8_regexp){1,256}/og) {
# 		$p = pos $_[0];
# 	}
# 	&check_error($err, $p == length $_[0], $E_MESSAGE_TEXT);
# }

require 'guestbook.pl';
