# # Guestbook
#
# This software is under Public Domain
#
# Requirements
# 	Perl 5.005 or later (maybe)
# 	CGI::Minimal (Tested with 1.29)
# 	CGI::FastTemplate (Tested with 1.09)
# 	Digest::MD5
# 	AtExit

use strict;
use warnings;

use CGI::Minimal;
use CGI::FastTemplate;
use AtExit;

require 'renlock.pl';

# constraint
#
if ($C_MIN_SIZE < 1) { $C_MIN_SIZE = 1; }
if ($C_MAX_PAGE < 1) { $C_MAX_PAGE = 32; }
if ($C_MAX_MESSAGES < 1) { $C_MAX_MESSAGES = 1; }
$C_MESSAGE_DIR =~ s|/+$||;
$C_MESSAGE_URL =~ s|/+$||;
for (@C_KEEP_EXTENSIONS) { $_ = '.'.$_; }

CGI::Minimal::reset_globals;
CGI::Minimal::max_read_size($C_MAX_MESSAGE_SIZE + $C_MAX_ATTACHMENT_SIZE + 2048);

binmode(STDIN);
binmode(STDOUT);

&dispatch;

sub dispatch {
	my $q = new CGI::Minimal;
	if ($q->truncated()) {
		&message_error($q, $E_REQUEST);
	} elsif (defined scalar $q->param($P_MESSAGE) ||
	         defined scalar $q->param($P_ATTACH)) {
		&doPost($q);
	} elsif (defined scalar $q->param($P_PAGE)) {
		&doList($q);
	} elsif (defined scalar $q->param($P_REBUILD)) {
		&doRebuild($q);
	} else {
		&doForm($q);
	}
}

sub doPost {
	my ($q) = @_;
	$ENV{'REQUEST_METHOD'} eq 'POST' or &message_error($q, $E_REQUEST);
	my ($files, @error);
	{
		my $ovw = &check_overwrite_name($q, \@error);
		$files = &make_message_list($q, \@error);
		&check_overwrite_size($q, \@error, $ovw, $files);
	}
	my $ov = &make_overwrite_list($q);
	my $lock = &do_lock($q, @error);
	my $tmpname = "${C_LISTFILE}.tmp";
	{
		local *WFH;
		open(WFH, ">$tmpname")
			or &message_error($q, $E_WRITE_LIST, @error);
		binmode(WFH);
		unless (&write_new_names($files)) {
			close(WFH);
			&message_error($q, $E_WRITE_LIST, @error);
		}
		{
			local *RFH;
			if (open(RFH, $C_LISTFILE)) {
				binmode(RFH);
				my $count;
				while (<RFH>) {
					chomp;
					next unless (/$RE_FILENAME/o);
					next if (exists $ov->{$_});
					$count ++;
					if ($count > $C_MAX_MESSAGES) {
						unlink "$C_MESSAGE_DIR/$_";
						unlink map { chomp; "$C_MESSAGE_DIR/$_"; } <RFH>;
						last;
					}
					unless (print WFH $_, "\n") {
						close(RFH); close(WFH);
						&message_error($q, $E_WRITE_LIST, @error);
					}
				}
				close(RFH);
			}
		}
		close(WFH);
	}
	unlink map { "$C_MESSAGE_DIR/$_"; } keys(%$ov);
	undef $ov;
	rename($tmpname, $C_LISTFILE)
		or &message_error($q, $E_WRITE_LIST, @error);
	my @success;
	for (@$files) {
		if (&write_body("$C_MESSAGE_DIR/$_->[0]", ${$_->[1]})) {
			push(@success, $_->[0]);
		} else {
			&push_error(\@error, $E_WRITE_MESSAGE.$_->[0]);
		}
	}
	undef $files;
	&do_unlock($lock);
	if (@success) {
		&message_success($q, \@success, \@error);
	} else {
		&message_error($q, @error);
	}
}

sub check_overwrite_name {
	my ($q, $e) = @_;
	my @ov = sort { $b <=> $a; } map {
		($_ =~ /$RE_FILESIZE/o && -w "$C_MESSAGE_DIR/$_")
			or &message_error($q, $E_OVERWRITE_NAME, @$e);
		int($1);
	} grep(length, $q->param($P_OVERWRITE));
	\@ov;
}

sub check_overwrite_size {
	my ($q, $err, $ov, $files) = @_;
	my @sz = sort { $a <=> $b; } map { length(${$_->[1]}); } @$files;
	for (@$ov) {
		if (9 * $_ / 10 > pop(@sz)) {
			&message_error($q, $E_OVERWRITE_SIZE, @$err);
		}
	}
}

sub make_message_list {
	my ($q, $e) = @_;
	my @files;
	{
		my ($name, %before);
		foreach ($q->param($P_MESSAGE)) {
			(
			length $_
			&&
			&check_error($e, length $_ >= $C_MIN_SIZE, $E_MESSAGE_SHORT)
			&&
			&check_error($e, length $_ <= $C_MAX_MESSAGE_SIZE, $E_MESSAGE_LONG)
			&&
			&check_message_text($e, $_)
			) or next;
			$name = &generate_filename($_);
			(! exists $before{$name} && &check_message_exists($e, $name))
				or next;
			push(@files, ["$name.txt", \$_]);
			$before{$name} = 1;
		}
		$q->delete($P_MESSAGE);
		my @org = $q->param_filename($P_ATTACH);
		foreach ($q->param($P_ATTACH)) {
			(
			length $_
			&&
			&check_error($e, length $_ >= $C_MIN_SIZE, $E_MESSAGE_SHORT)
			&&
			&check_error($e, length $_ <= $C_MAX_ATTACHMENT_SIZE, $E_MESSAGE_LONG)
			) or next;
			$name = &generate_filename($_);
			(! exists $before{$name} && &check_message_exists($e, $name))
				or next;
			push(@files, [$name.&select_extention(shift(@org)), \$_]);
			$before{$name} = 1;
		}
		$q->delete($P_ATTACH);
	}
	@files or &message_error($q, $E_MESSAGE_NOME, @$e);
	\@files;
}

sub check_message_exists {
	my ($err, $name) = @_;
	foreach ('.txt', '.bin', @C_KEEP_EXTENSIONS) {
		&check_error($err, ! -e "$C_MESSAGE_DIR/$name$_", $E_MESSAGE_SAME)
			or return 0;
	}
	1;
}

sub select_extention {
	my $i;
	foreach $i (@C_KEEP_EXTENSIONS) {
		return $i if (substr($_[0], - length($i)) eq $i);
	}
	'.bin';
}

sub make_overwrite_list {
	my ($q) = @_;
	my %ov = map {$_ => 1; } grep(length, $q->param($P_OVERWRITE));
	$q->delete($P_OVERWRITE);
	\%ov;
}

sub write_new_names {
	my ($files) = @_;
	local $, = "\n"; local $\ = "\n";
	print WFH map { $_->[0]; } @$files;
}

sub write_body {
	local *FH;
	open(FH, '>'.$_[0]) or return 0;
	binmode(FH);
	unless (print FH $_[1]) {
		close(FH);
		return 0;
	}
	close(FH);
}

sub check_error {
	my $err = shift;
	my $expr = shift;
	push(@$err, @_) unless ($expr);
	$expr;
}

sub push_error {
	my $err = shift;
	push(@$err, @_);
}

sub doList {
	my ($q) = @_;
	my $page = int($q->param($P_PAGE)); 
	my $p = ($page - 1) * $C_MAX_PAGE;
	if ($p < 0 || $p > $C_MAX_MESSAGES) {
		$p = 0; $page = 1;
	}
	my $t = new CGI::FastTemplate();
	$t->define_nofile(main => $T_LIST,
	                  list => $T_LIST_ITEM,
	                  nav_prev => $T_NAV_PREV,
	                  nav_next => $T_NAV_NEXT);
	$t->assign(C_THIS_URL => $C_THIS_URL,
	           C_MESSAGE_URL => $C_MESSAGE_URL || $C_MESSAGE_DIR,
	           P_PAGE => $P_PAGE,
	           P_REBUILD => $P_REBUILD,
	           T_NAV_PREV => '',
	           T_NAV_NEXT => '');
	my $count = 0;
	my $lock = &do_lock($q);
	{
		local *FH;
		if (open(FH, $C_LISTFILE)) {
			binmode(FH);
			if ($page > 1) {
				$t->assign(PAGE => $page - 1);
				$t->parse(T_NAV_PREV => 'nav_prev');
				$t->clear('PAGE');
			}
			while (<FH>) {
				chomp;
				if (/$RE_FILENAME/o) {
					$count ++;
					if ($count > $p) {
						$t->assign(C_THIS_URL => $C_THIS_URL,
						           P_OVERWRITE => $P_OVERWRITE,
						           ITEM_NAME => $_);
						$t->parse(T_LIST_ITEM => '.list');
						$t->clear('ITEM_NAME');
					}
					last if ($count >= $p + $C_MAX_PAGE);
				}
			}
			unless (eof(FH)) {
				$t->assign(PAGE => $page + 1);
				$t->parse(T_NAV_NEXT => 'nav_next');
				$t->clear('PAGE');
			}
			close(FH);
		}
	}
	&do_unlock($lock);
	unless ($count > 0) {
		$t->define_nofile(none => $T_NO_ITEM);
		$t->parse(T_LIST_ITEM => 'none');
	}
	$t->parse(MAIN => 'main');
	&print_header();
	$t->print();
	$t->clear();
}

sub doRebuild {
	my ($q) = @_;
	my $lock = &do_lock($q);
	if (-e $C_LISTFILE) {
		local *FH;
		if (open(FH, "$C_REBUILD_FILE")) {
			binmode(FH);
			my $before = int(<FH>);
			close(FH);
			time() - $before > $C_REBUILD_INTERVAL
				or &message_error($q, $E_REBUILD_INTERVAL);
		}
	}
	{
		my @list;
		{
			local *DH;
			opendir(DH, $C_MESSAGE_DIR)
				or &message_error($q, $E_REBUILD_DIRECTORY);
			@list = map { $_->[1]; } sort { $b->[0] <=> $a->[0]; }
			map {
				my (undef, undef, undef, undef, undef, undef, undef, undef,
				    undef, $mtime, undef, undef, undef)
				    = stat("$C_MESSAGE_DIR/$_");
				[$mtime, $_];
			} grep { $_ =~ /$RE_FILENAME/o && -f "$C_MESSAGE_DIR/$_"; }
			readdir(DH);
			closedir(DH);
		}
		&write_list(\@list)
			or &message_error($q, $E_WRITE_LIST);
	}
	&write_lines($C_REBUILD_FILE, time())
		or &message_error($q, $E_REBUILD_FILE);
	&do_unlock($lock);
	&message_success($q);
}

sub write_list {
	&write_lines($C_LISTFILE, @{$_[0]});
}

sub write_lines {
	my $name = shift;
	my $tmpname;
	$tmpname = "${name}.tmp";
	{
		local *FH;
		open(FH, ">$tmpname") or return 0;
		binmode(FH);
		local $, = "\n"; local $\ = "\n";
		unless (print FH @_) {
			close(FH);
			return 0;
		}
		close(FH);
	}
	rename($tmpname, $name);
}

sub doForm {
	my ($q) = @_;
	my ($ov) = ($q->param($P_OVERWRITE));
	$ov = "" unless (defined $ov && $ov =~ /$RE_FILENAME/o);
	my $t = new CGI::FastTemplate();
	$t->define_nofile(main => $T_FORM);
	$t->assign(C_THIS_URL => $C_THIS_URL,
	           C_MIN_SIZE => $C_MIN_SIZE,
	           C_MAX_MESSAGE_SIZE => $C_MAX_MESSAGE_SIZE,
	           C_MAX_ATTACHMENT_SIZE => $C_MAX_ATTACHMENT_SIZE,
	           C_MAX_MESSAGES => $C_MAX_MESSAGES,
	           C_LIST_URL => $C_LIST_URL || $C_LISTFILE,
	           C_MESSAGE_URL => $C_MESSAGE_URL || $C_MESSAGE_DIR,
	           P_MESSAGE => $P_MESSAGE,
	           P_ATTACH => $P_ATTACH,
	           P_OVERWRITE => $P_OVERWRITE,
	           P_PAGE => $P_PAGE,
	           P_REBUILD => $P_REBUILD,
	           EXTENSIONS => join(' ', @C_KEEP_EXTENSIONS),
	           OVERWRITE => $q->htmlize($ov));
	$t->parse(MAIN => 'main');
	&print_header();
	$t->print();
	$t->clear();
}

sub message_success {
	my ($q, $list, $error) = @_;
	my $t = new CGI::FastTemplate();
	$t->define_nofile(main => $T_STATUS,
	                  message => $T_MESSAGE,
	                  list => $T_STATUS_ITEM,
	                  nav => $T_NAV_LIST);
	$t->assign(STATUS => 'Success.',
	           C_THIS_URL => $C_THIS_URL,
	           C_MESSAGE_URL => $C_MESSAGE_URL || $C_MESSAGE_DIR,
	           P_PAGE => $P_PAGE);
	$t->assign(T_MESSAGE => '');
	for (@$list) {
		$t->assign(ITEM_NAME => $_);
		$t->parse(T_MESSAGE => '.list');
		$t->clear('ITEM_NAME');
	}
	for (@$error) {
		$t->assign(REASON => $q->htmlize($_));
		$t->parse(T_MESSAGE => '.message');
		$t->clear('REASON');
	}
	$t->parse(T_NAV_LIST => 'nav');
	$t->parse(MAIN => 'main');
	&print_header();
	$t->print();
	$t->clear();
	exit;
}

sub message_error {
	my $q = shift;
	my $t = new CGI::FastTemplate();
	$t->define_nofile(main => $T_STATUS, message => $T_MESSAGE);
	$t->assign(T_MESSAGE => '');
	for (@_) {
		$t->assign(REASON => $q->htmlize($_));
		$t->parse(T_MESSAGE => '.message');
		$t->clear('REASON');
	}
	$t->assign(STATUS => 'ERROR.',
	           T_NAV_LIST => '',
	           C_THIS_URL => $C_THIS_URL);
	$t->parse(MAIN => 'main');
	&print_header();
	$t->print();
	$t->clear();
	exit;
}

sub print_header {
	print
		"Content-Type: text/html\r\n",
		"Cache-Control: no-cache\r\n",
		"Pragma: no-cache\r\n\r\n";
}

sub do_lock {
	my $q = shift;
	my $lh = &lock(dir => $C_LOCKDIR, basename => $C_LOCKFILE);
	&message_error($q, $E_LOCK, @_) unless $lh;
	my $lock = AtExit->new(\&unlock, $lh);
	[$lock, $lh];
}

sub do_unlock {
	&unlock($_[0]->[1]);
}

1;
